/*
___________________________________________________________________________________________________

   __  __ _______ _____                  _                                      
  |  \/  |__   __|  __ \                | |                                     
  | \  / |  | |  | |__) | __ _ _ __   __| | ___  _ __ ___       ___ _ __  _ __  
  | |\/| |  | |  |  _  / / _` | '_ \ / _` |/ _ \| '_ ` _ \     / __| '_ \| '_ \ 
  | |  | |  | |  | | \ \| (_| | | | | (_| | (_) | | | | | | _ | (__| |_) | |_) |
  |_|  |_|  |_|  |_|  \_\\__,_|_| |_|\__,_|\___/|_| |_| |_|(_) \___| .__/| .__/ 
                                                                   | |   | |    
                                                                   |_|   |_|    

___________________________________________________________________________________________________

  Generador de numeros aleatorios Mersenne Twister. 
___________________________________________________________________________________________________

  2012 Javier V.R.
___________________________________________________________________________________________________ 
*/ 

#include "mtrandom.h"

// ------------------------------------------------------------------------------------------------


namespace JVLib 
{ 

  /*------------------------------------------------------------
    Constructor. 
  ------------------------------------------------------------*/ 
  CMTRandom::CMTRandom() : 
                       m_next(nullptr), 
                       m_init_seed(0), 
                       m_left(0) 
  { 
    static UI32 seed_inc = 0; 
    SetSeed(((UI32)time(nullptr)) + seed_inc++); 
  } 


  /*------------------------------------------------------------
    Constructor. 
  ------------------------------------------------------------*/ 
  CMTRandom::CMTRandom(UI32 seed) : 
                         m_next(nullptr), 
                         m_init_seed(0), 
                         m_left(0) 
  { 
    SetSeed(seed); 
  } 


  /*------------------------------------------------------------
    Destructor. 
  ------------------------------------------------------------*/ 
  CMTRandom::~CMTRandom() 
  { 
    
  } 


  /*------------------------------------------------------------
    SetSeed. 
  ------------------------------------------------------------*/ 
  void CMTRandom::SetSeed(UI32 seed) 
  { 
    m_init_seed = seed; 
    m_left      = 0;
    
    register UI32  x = (seed | 1U) & 0xFFFFFFFFU; 
    register UI32 *s = m_state;
    register I32   j; 

    for(*s++=x, j=N; --j; *s++ = (x*=69069U) & 0xFFFFFFFFU); 
  } 


  /*------------------------------------------------------------
    Reload. 
  ------------------------------------------------------------*/
  UI32 CMTRandom::Reload() 
  { 
    register UI32 *p0 = m_state; 
    register UI32 *p2 = m_state + 2; 
    register UI32 *pM = m_state + M; 
    register UI32  s0; 
    register UI32  s1;
    register I32   j;

    if(m_left < -1)
    {
      SetSeed(m_init_seed); 
    } 

    m_left = N - 1; 
    m_next = m_state + 1;

    for(s0=m_state[0], s1=m_state[1], j=N - M + 1; --j; s0=s1, s1=*p2++) 
    {
      *p0++ = *pM++ ^ (MixBits(s0, s1) >> 1) ^ (LoBit(s1) ? K : 0U);
    }

    for(pM=m_state, j=M; --j; s0=s1, s1=*p2++) 
    {
      *p0++ = *pM++ ^ (MixBits(s0, s1) >> 1) ^ (LoBit(s1) ? K : 0U);
    }

    s1  = m_state[0]; 
    *p0 = *pM ^ (MixBits(s0, s1) >> 1) ^ (LoBit(s1) ? K : 0U); 
    s1 ^= (s1 >> 11);
    s1 ^= (s1 << 7) & 0x9D2C5680U;
    s1 ^= (s1 << 15) & 0xEFC60000U; 
    
    return(s1 ^ (s1 >> 18));
  } 


  /*------------------------------------------------------------
    GetInt. 
  ------------------------------------------------------------*/
  Int CMTRandom::GetInt(Int min, 
                        Int max, 
                        bool force_i32) 
  { 
    JVLIB_ASSERT(min <= max); 

    #if defined(JVLIB_OS_32_BITS) 
      return((GetRaw() % ((max + 1) - min)) + min); 
    #elif defined(JVLIB_OS_64_BITS) 
      if(force_i32) 
      { 
        // Forzar su calculo a 32 bits 
        return((GetRaw() % ((max + 1) - min)) + min); 
      } 
      UI64 rnd = (UI64(GetRaw()) << 32) + UI64(GetRaw()); 

      return((rnd % ((max + 1) - min)) + min);
    #else 
      #error Define el modo de compilacion (32 o 64 bits)! 
    #endif 
  } 


  /*------------------------------------------------------------
    GetUInt. 
  ------------------------------------------------------------*/
  UInt CMTRandom::GetUInt(UInt min, 
                          UInt max, 
                          bool force_ui32) 
  { 
    JVLIB_ASSERT(min <= max); 

    #if defined(JVLIB_OS_32_BITS) 
      return((GetRaw() % ((max + 1) - min)) + min); 
    #elif defined(JVLIB_OS_64_BITS) 
      if(force_ui32) 
      { 
        // Forzar su calculo a 32 bits 
        return((GetRaw() % ((max + 1) - min)) + min); 
      } 
      UI64 rnd = (UI64(GetRaw()) << 32) + UI64(GetRaw()); 

      return((rnd % ((max + 1) - min)) + min);
    #else 
      #error Define el modo de compilacion (32 o 64 bits)! 
    #endif 
  } 

  
  /*------------------------------------------------------------
    GetI32. 
  ------------------------------------------------------------*/
  I32 CMTRandom::GetI32(I32 min, 
                        I32 max) 
  { 
    JVLIB_ASSERT(min <= max); 

    return((GetRaw() % ((max + 1) - min)) + min);
  } 


  /*------------------------------------------------------------
    GetUI32. 
  ------------------------------------------------------------*/
  UI32 CMTRandom::GetUI32(UI32 min, 
                          UI32 max) 
  { 
    JVLIB_ASSERT(min <= max); 

    return((GetRaw() % ((max + 1) - min)) + min);
  } 


  /*------------------------------------------------------------
    GetI64. 
  ------------------------------------------------------------*/
  I64 CMTRandom::GetI64(I64 min, 
                        I64 max) 
  { 
    JVLIB_ASSERT(min <= max); 

    UI64 rnd = (UI64(GetRaw()) << 32) + UI64(GetRaw()); 

    return((rnd % ((max + 1) - min)) + min);
  } 


  /*------------------------------------------------------------
    GetUI64. 
  ------------------------------------------------------------*/
  UI64 CMTRandom::GetUI64(UI64 min, 
                          UI64 max) 
  { 
    JVLIB_ASSERT(min <= max); 

    UI64 rnd = (UI64(GetRaw()) << 32) + UI64(GetRaw()); 

    return((rnd % ((max + 1) - min)) + min);
  } 


  /*------------------------------------------------------------
    GetFloat. 
  ------------------------------------------------------------*/
  float CMTRandom::GetFloat(float min, 
                            float max) 
  { 
    JVLIB_ASSERT(min <= max); 

    static const float c_rand_max_inv = 1.0 / float(JVLIB_UI32_MAX); 

    return((float(GetRaw()) * c_rand_max_inv) * (max - min) + min); 
  } 


  /*------------------------------------------------------------
    GetDouble. 
  ------------------------------------------------------------*/
  double CMTRandom::GetDouble(double min, 
                              double max) 
  { 
    JVLIB_ASSERT(min <= max); 

    static const double c_rand_max_inv = 1.0 / double(JVLIB_UI32_MAX); 

    return((double(GetRaw()) * c_rand_max_inv) * (max - min) + min); 
  } 


  /*------------------------------------------------------------
    GetDoubleExtra. 
  ------------------------------------------------------------*/
  double CMTRandom::GetDoubleExtra(double min, 
                                   double max) 
  { 
    JVLIB_ASSERT(min <= max); 

    static const double c_rand_max_inv = 1.0 / double(JVLIB_UI64_MAX); 

    return((double((UI64(GetRaw()) << 32) + UI64(GetRaw())) * c_rand_max_inv) * (max - min) + min); 
  } 

} // namespace JVLib 
