/*
___________________________________________________________________________________________________

   __  __ _______ _____                  _                     _     
  |  \/  |__   __|  __ \                | |                   | |    
  | \  / |  | |  | |__) | __ _ _ __   __| | ___  _ __ ___     | |__  
  | |\/| |  | |  |  _  / / _` | '_ \ / _` |/ _ \| '_ ` _ \    | '_ \ 
  | |  | |  | |  | | \ \| (_| | | | | (_| | (_) | | | | | | _ | | | |
  |_|  |_|  |_|  |_|  \_\\__,_|_| |_|\__,_|\___/|_| |_| |_|(_)|_| |_|
                                                                     
                                                                     

___________________________________________________________________________________________________

  Generador de numeros aleatorios Mersenne Twister. 
  Genera una distribucion uniforme de enteros en el rango [0 .. 2^32 - 1]. 
  
  Tiempos de ejecucion comparativos con rand(): 
    - 1.000.000.000 de UI32 : rand() = 16.3 seg :: CMTRandom = 5.5 seg. 

  - No heredable. 
  - No copiable. 
___________________________________________________________________________________________________

  2012 Javier V.R.
___________________________________________________________________________________________________ 
*/ 

#ifndef __MTRANDOM_H__0488813C__
#define __MTRANDOM_H__0488813C__

#include <jvlib/core.h> 
#include <jvlib/math.h> 

#include <ctime> 

// ------------------------------------------------------------------------------------------------


namespace JVLib 
{ 

  class CMTRandom 
  { 
    /*------------------------------------------------------------
      ::::: Constantes y variables estaticas ::::: 
    ------------------------------------------------------------*/ 
    static const I32  N = 624; 
    static const I32  M = 397; 
    static const UI32 K = 0x9908B0DFU; 

    
    /*------------------------------------------------------------
      ::::: Variables miembro privadas ::::: 
    ------------------------------------------------------------*/ 
    UI32  m_state[N + 1]; 
    UI32 *m_next; 
    UI32  m_init_seed; 
    I32   m_left; 


    /*------------------------------------------------------------
      Impedir la copia y asignacion. 
    ------------------------------------------------------------*/
    CMTRandom(const CMTRandom &); 
    CMTRandom &operator=(const CMTRandom &); 

    
    /*------------------------------------------------------------
      Enmascarar los bits bajos de <u>. 
    ------------------------------------------------------------*/
    inline UI32 HiBit(UI32 u) const 
    { 
      return(u & 0x80000000U); 
    }

    
    /*------------------------------------------------------------
      Enmascarar los bits altos de <u>. 
    ------------------------------------------------------------*/
    inline UI32 LoBit(UI32 u) const 
    { 
      return(u & 0x00000001U); 
    }

    
    /*------------------------------------------------------------
      Enmascarar el bit mas alto de <u>. 
    ------------------------------------------------------------*/
    inline UI32 LoBits(UI32 u) const 
    { 
      return(u & 0x7FFFFFFFU); 
    }

    
    /*------------------------------------------------------------
      Mover los bits altos de <u> a la parte alta de <v>. 
    ------------------------------------------------------------*/
    inline UI32 MixBits(UI32 u, UI32 v) const 
    {
      return(HiBit(u) | LoBits(v)); 
    }
    
    
    /*------------------------------------------------------------
      Recargar secuencia. 
    ------------------------------------------------------------*/
    UI32 Reload(); 
    

  public: 
    /*------------------------------------------------------------
      Constructor por defecto. 
      Inicializa el generador con la hora actual mas un contador 
      estatico interno para evitar que distintos objetos 
      CMTRandom creados simultaneamente tengan la misma semilla. 
    ------------------------------------------------------------*/ 
    CMTRandom(); 


    /*------------------------------------------------------------
      Constructor. 
      Inicializa el generador con la semilla <seed> indicada. 
    ------------------------------------------------------------*/ 
    CMTRandom(UI32 seed); 


    /*------------------------------------------------------------
      Destructor. 
    ------------------------------------------------------------*/ 
    ~CMTRandom(); 


    /*------------------------------------------------------------
      Fijar la semilla del generador. 
    ------------------------------------------------------------*/
    void SetSeed(UI32 seed);


    /*------------------------------------------------------------
      Obtiene un numero aleatorio UI32 de (0 .. 2^32 - 1). 
    ------------------------------------------------------------*/
    inline UI32 GetRaw() 
    {
      if(--m_left < 0) 
      { 
        return(Reload()); 
      } 

      register UI32 y = *m_next++;
      y ^= (y >> 11);
      y ^= (y <<  7) & 0x9D2C5680U;
      y ^= (y << 15) & 0xEFC60000U; 
      
      return(y ^ (y >> 18));
    }  


    /*------------------------------------------------------------
      Obtiene un numero aleatorio I32 o I64 entre <min> y <max> 
      ambos inclusive. 
      Si <force_i32> es true se calculara siempre como I32. 
    ------------------------------------------------------------*/
    Int GetInt(Int min, Int max, bool force_i32=false); 

    
    /*------------------------------------------------------------
      Obtiene un numero aleatorio UI32 o UI64 entre <min> y <max> 
      ambos inclusive. 
      Si <force_ui32> es true se calculara siempre como UI32. 
    ------------------------------------------------------------*/
    UInt GetUInt(UInt min, UInt max, bool force_ui32=false); 
    
    
    /*------------------------------------------------------------
      Obtiene un numero aleatorio I32 entre <min> y <max> ambos 
      inclusive. 
    ------------------------------------------------------------*/
    I32 GetI32(I32 min, I32 max); 


    /*------------------------------------------------------------
      Obtiene un numero aleatorio UI32 entre <min> y <max> ambos 
      inclusive. 
    ------------------------------------------------------------*/
    UI32 GetUI32(UI32 min, UI32 max); 


    /*------------------------------------------------------------
      Obtiene un numero aleatorio I64 entre <min> y <max> ambos 
      inclusive. 
    ------------------------------------------------------------*/
    I64 GetI64(I64 min, I64 max); 


    /*------------------------------------------------------------
      Obtiene un numero aleatorio UI64 entre <min> y <max> ambos 
      inclusive. 
    ------------------------------------------------------------*/
    UI64 GetUI64(UI64 min, UI64 max); 


    /*------------------------------------------------------------
      Obtiene un numero aleatorio float entre <min> y <max> 
      ambos inclusive. 
    ------------------------------------------------------------*/
    float GetFloat(float min, float max); 


    /*------------------------------------------------------------
      Obtiene un numero aleatorio double entre <min> y <max> 
      ambos inclusive. 
    ------------------------------------------------------------*/
    double GetDouble(double min, double max); 


    /*------------------------------------------------------------
      Obtiene un numero aleatorio double entre <min> y <max> 
      ambos inclusive, con una presicion interna de 64 bits. 
    ------------------------------------------------------------*/
    double GetDoubleExtra(double min, double max); 
    
  }; 

} // namespace JVLib 

#endif // __MTRANDOM_H__0488813C__
